// ======================================================================
//
// material.cpp
// jeff grills
//
// copyright 1998 origin systems inc
//
// -qq- support materials using other materials
//
// ======================================================================

#include "First.h"
#include "Material.h"

#include "ExitChain.h"
#include "Gl.h"
#include "Iff.h"
#include "ConfigEngine.h"

#include <stdio.h>
#include <stdlib.h>

// ======================================================================

const Tag TAG_ALPH = TAG(A,L,P,H);
const Tag TAG_SALF = TAG(S,A,L,F);
const Tag TAG_SMAT = TAG(S,M,A,T);
const Tag TAG_AMAT = TAG(A,M,A,T);
const Tag TAG_PXL8 = TAG(P,X,L,8);
const Tag TAG_PX32 = TAG(P,X,3,2);
const Tag TAG_MMAP = TAG(M,M,A,P);

// ======================================================================

Material                 *MaterialList::materialList;
const Material           *MaterialList::debug;
const StaticMaterial     *MaterialList::redMaterial;
const StaticMaterial     *MaterialList::greenMaterial;
const StaticMaterial     *MaterialList::blueMaterial;
const StaticMaterial     *MaterialList::blackMaterial;
const StaticMaterial     *MaterialList::whiteMaterial;
int                       MaterialList::materialCount;
real                      MaterialList::time;

int                       Material::globalToken;

// ======================================================================

void MaterialList::install(void)
{
	ExitChain::add(&MaterialList::remove, "MaterialList", 0);
}

// ----------------------------------------------------------------------

void MaterialList::remove(void)
{
	release(debug);
	debug = NULL;

	release(redMaterial);
	redMaterial = NULL;

	release(blueMaterial);
	blueMaterial = NULL;

	release(greenMaterial);
	greenMaterial = NULL;

	release(blackMaterial);
	blackMaterial = NULL;

	release(whiteMaterial);
	whiteMaterial = NULL;
}

// ----------------------------------------------------------------------

const Material *MaterialList::debugMaterial(void)
{
	if (!debug)
		debug = fetch(ConfigEngine::debugMaterial());

	return debug;
}

// ----------------------------------------------------------------------

const StaticMaterial *MaterialList::red(void)
{
	if (!redMaterial)
		redMaterial = fetch(255, 0, 0);

	return redMaterial;
}

// ----------------------------------------------------------------------

const StaticMaterial *MaterialList::green(void)
{
	if (!greenMaterial)
		greenMaterial = fetch(0, 255, 0);

	return greenMaterial;
}

// ----------------------------------------------------------------------

const StaticMaterial *MaterialList::blue(void)
{
	if (!blueMaterial)
		blueMaterial = fetch(0, 0, 255);

	return blueMaterial;
}

// ----------------------------------------------------------------------

const StaticMaterial *MaterialList::black(void)
{
	if (!blackMaterial)
		blackMaterial = fetch(0, 0, 0);

	return blackMaterial;
}

// ----------------------------------------------------------------------

const StaticMaterial *MaterialList::white(void)
{
	if (!whiteMaterial)
		whiteMaterial = fetch(255, 255, 255);

	return whiteMaterial;
}

// ----------------------------------------------------------------------

void MaterialList::debugDump(void)
{
	DPRINT(("MaterialList  %4d=exist\n", materialCount));
}

// ----------------------------------------------------------------------

const Material *MaterialList::fetch(const Material *mat)
{
	Material *m;

	if (mat)
	{
		// copy the material
		m = mat->copy();

		// if copying created a new material, put the new material on the material list
		if (m != mat)
		{
			m->next = materialList;
			materialList = m;
		}
	}
	else
		m = NULL;

	return m;
}

// ----------------------------------------------------------------------

const Material *MaterialList::fetch(int num)
{
	if ((num & MAT_MASK) == MAT_RGB)
	{
		int r, g, b;

		r = (num >> 16) & 0xff;
		g = (num >>  8) & 0xff;
		b = (num >>  0) & 0xff;
		return fetch(r, g, b);
	}

	char buffer[256];
	sprintf(buffer, "mat\\%08d.mat", num);

	return fetch(buffer);
}

// ----------------------------------------------------------------------

const StaticMaterial *MaterialList::fetch(int r, int g, int b)
{
	char buffer[256];
	sprintf(buffer, "RGB %d %d %d", r, g, b);

	return fetch(buffer)->getStaticMaterial();
}

// ----------------------------------------------------------------------

const Material *MaterialList::fetch(const char *name)
{
	Material *m;
	CrcString crcname(name);

	// search for the material
	for (m = materialList; m && (!m->getName() || crcname.compareCrcString(*(m->getCrcName())) != 0); m = m->next)
		;

	// load the material and stick it in the list
	if (m)
		return fetch(m);

	if (strncmp(name, "RGB ", 4) == 0)
	{
		m = NEW StaticMaterial(name, NULL);
	}
	else
	{
		Iff iff(name);

		switch (iff.currentName())
		{
			case TAG_SALF:
			case TAG_BITM:
			case TAG_SMAT:
				m = NEW StaticMaterial(name, &iff);
				break;

			case TAG_ANIM:
			case TAG_AMAT:
				m = NEW AnimatingMaterial(name, &iff);
				break;

			default:
				DEBUG_FATAL(true, ("MaterialList::fetch bad type"));
				m = NULL;
				break;
		}
	}

	m->next = materialList;
	materialList = m;

	return m;
}

// ----------------------------------------------------------------------

const Material *MaterialList::fetch(Iff *iff)
{
	Material *m;
	char buffer[256];

	switch (iff->currentName())
	{
		case TAG_NAME:
			iff->enterChunk();
			iff->read_string(buffer, sizeof(buffer));
			iff->exitChunk();
			return MaterialList::fetch(buffer);
			break;

		case TAG_SMAT:
			m = NEW StaticMaterial(NULL, iff);
			break;

		case TAG_AMAT:
			m = NEW AnimatingMaterial(NULL, iff);
			break;

		default:
			iff->formatCurrentName(buffer, sizeof(buffer));
			DEBUG_FATAL(true, ("MaterialList::fetch(iff) unknown tag %s", buffer));
			m = NULL;
			break;
	}

	m->next = materialList;
	materialList = m;

	return m;
}

// ----------------------------------------------------------------------

void MaterialList::release(const Material *mat)
{
	Material *back, *front;

	// decrement the users and possibly free the material
	if (mat && --mat->usage == 0)
	{
		// search list
		for (back = NULL, front = materialList; front && front != mat; back = front, front = front->next)
			;

		if (front)
		{
			if (back)
				back->next = front->next;
			else
				materialList = front->next;

			delete mat;
		}
		else
			DEBUG_FATAL(true, ("MaterialList::release not found"));
	}
}

// ----------------------------------------------------------------------

void MaterialList::createAllData(void)
{
	Material *m;

	for (m = materialList; m; m = m->next)
		if (m->getId() == MaterialIdStatic)
			static_cast<StaticMaterial *>(m)->createData();
}

// ----------------------------------------------------------------------

void MaterialList::destroyAllData(void)
{
	Material *m;

	for (m = materialList; m; m = m->next)
		if (m->getId() == MaterialIdStatic)
			static_cast<StaticMaterial *>(m)->destroyData();
}


// ======================================================================

Material::Material(const char *newName, MaterialId newId)
{
	name  = MEM_OWN(strdup(newName));
	crcName  = NEW CrcString(name);
	id    = newId;

	usage = 1;
	token = ++globalToken;
	next  = NULL;
}

// ----------------------------------------------------------------------

Material::Material(const Material &m)
{
	// copy the base material information
	name   = MEM_OWN(strdup(m.name));
	crcName = NEW CrcString(name);
	id     = m.id;

	usage  = 1;
	token  = ++globalToken;
	next   = NULL;
}

// ----------------------------------------------------------------------

Material::~Material(void)
{
	Gl::materialDestroyed(this);

	delete crcName;
	delete [] name;
	next = NULL;
}

// ----------------------------------------------------------------------

Material *Material::copy(void) const
{
	++usage;
	return const_cast<Material *>(this);
}

// ======================================================================

StaticMaterial::StaticMaterial(const char *newName, Iff *iff)
: Material(newName, MaterialIdStatic)
{
	init();
	load(iff);
	createData();
}

// ----------------------------------------------------------------------

StaticMaterial::~StaticMaterial(void)
{
	int i;

	destroyData();

	if (rgbMap8)
	{
		for (i = 0; i < mipMapLevels; ++i)
			delete [] rgbMap8[i];

		delete [] rgbMap8;
		rgbMap8 = NULL;

		if (alphaMap)
			for (i = 0; i < mipMapLevels; ++i)
				delete [] alphaMap[i];

		delete [] alphaMap;
		alphaMap = NULL;

		PaletteList::release(palette);
		palette = NULL;
	}

	if (rgbMap32)
	{
		for (i = 0; i < mipMapLevels; ++i)
			delete [] rgbMap32[i];

		delete [] rgbMap32;
		rgbMap32 = NULL;
	}
}

// ----------------------------------------------------------------------

void StaticMaterial::init(void)
{
	data          = NULL;

	r             = 0;
	g             = 0;
	b             = 0;
	a             = 255;

	width         = 0;
	height        = 0;
	mipMapLevels  = 0;

	palette       = NULL;
	rgbMap8       = NULL;
	alphaMap      = NULL;

	rgbMap32      = NULL;

	chromakey     = ChromakeyOff;
	alpha         = AlphaOff;
	color         = ColorConstant;
}

// ----------------------------------------------------------------------

void StaticMaterial::load(Iff *iff)
{
	if (iff == NULL)
	{
		// handle constant colored materials
		const char sep[] = " ";
		char *temp = MEM_OWN(strdup(name));

		strtok(temp, sep);
		r = static_cast<u_int8>(atoi(strtok(NULL, sep)));
		g = static_cast<u_int8>(atoi(strtok(NULL, sep)));
		b = static_cast<u_int8>(atoi(strtok(NULL, sep)));

		delete [] temp;
	}
	else
	{
		// figure out what version to load
		switch (iff->currentName())
		{
			case TAG_BITM:
				load_BITM(iff);
				break;

			case TAG_SALF:
				load_SALF(iff);
				break;

			case TAG_SMAT:
				iff->enterForm();
				switch (iff->currentName())
				{
					case TAG_0001:
						iff->seekForm(TAG_0001);
							load_0001(iff);
						iff->exitForm();
						break;

					default:
						DEBUG_FATAL(true, ("StaticMaterial::load bad version"));
						break;
				}
				iff->exitForm();
				break;

			default:
				DEBUG_FATAL(true, ("StaticMaterial::load unknown type"));
				break;
		}
	}
}

// ----------------------------------------------------------------------

void StaticMaterial::load_0001(Iff *iff)
{
	int i, flags;

	iff->seekChunk(TAG_INFO);
		width         = iff->read_int32();
		height        = iff->read_int32();
		mipMapLevels  = iff->read_int32();
		flags         = iff->read_int32();
	iff->exitChunk();

	// transparency
	if (flags & BINARY1(0001))
		chromakey = ChromakeyOn;

	// constant RGB
	if ((flags & BINARY1(0100)) == 0)
		color = ColorMap;

	// translucency
	if (flags & BINARY1(0010))
		// constant alpha
		if (flags & BINARY1(1000))
			alpha = AlphaConstant;
		else
			alpha = AlphaMap;

	// optionally load solid color information
	if (iff->seekChunk(TAG_COLR, Iff::FlagOptional))
	{
		r = iff->read_u_int8();
		g = iff->read_u_int8();
		b = iff->read_u_int8();
		a = static_cast<u_int8>(255 - iff->read_u_int8());

		if (a != 255)
			alpha = AlphaConstant;

		iff->exitChunk();
	}

	// optionally load palette
	if (iff->seekForm(TAG_PAL, Iff::FlagOptional))
	{
		palette = PaletteList::fetch(*iff);
		iff->exitForm();
	}

	bool hasPixel8, firstHasPixel8, hasAlpha8, firstHasAlpha8, hasPixel32, firstHasPixel32;

	firstHasPixel8  = false;
	firstHasAlpha8  = false;
	firstHasPixel32 = false;

	// load all the maps
	for (i = 0; i < mipMapLevels; ++i)
	{
		iff->seekForm(TAG_MMAP);

			// load the palettized texture map information
			if (iff->seekChunk(TAG_PXL8, Iff::FlagOptional))
			{
				DEBUG_FATAL(!palette, ("StaticMaterial::load_0001 rgbMap8 but no palette"));

				hasPixel8 = true;
				color = ColorMap;

				if (!rgbMap8)
				{
					rgbMap8 = NEW Pixel8 *[mipMapLevels];
					memset(rgbMap8, 0, sizeof(*rgbMap8) * mipMapLevels);
				}

				DEBUG_FATAL(iff->chunkLength(sizeof(Pixel8)) != ((width >> i) * (height >> i)), ("StaticMaterial::load_0001 pxl8 bad size"));
				rgbMap8[i] = MEM_OWN(iff->read_pixel8(Iff::ReadRest));

				iff->exitChunk();
			}
			else
			{
				hasPixel8 = false;
			}

			if (i == 0)
				firstHasPixel8 = hasPixel8;
			else
				DEBUG_FATAL(hasPixel8 != firstHasPixel8, ("StaticMaterial::load_0001 some mip map levels have pxl8, some don't"));

			// load the optional alpha channel
			if (iff->seekChunk(TAG_ALPH, Iff::FlagOptional))
			{
				DEBUG_FATAL(alpha != AlphaMap, ("StaticMaterial::load_0001 alpha but not translucency"));

				hasAlpha8 = true;
				alpha = AlphaMap;

				if (!alphaMap)
				{
					alphaMap = NEW Pixel8 *[mipMapLevels];
					memset(alphaMap, 0, sizeof(*alphaMap) * mipMapLevels);
				}

				DEBUG_FATAL(iff->chunkLength(sizeof(Pixel8)) != ((width >> i) * (height >> i)), ("StaticMaterial::load_0001 alph bad size"));
				alphaMap[i] = MEM_OWN(iff->read_pixel8(Iff::ReadRest));

				{
					int j;
					for (j = 0; j < (((width >> i) * (height >> i))); ++j)
						alphaMap[i][j] = static_cast<u_int8>(255 - alphaMap[i][j]);
				}

				iff->exitChunk();
			}
			else
			{
				hasAlpha8 = false;
			}

			if (i == 0)
				firstHasAlpha8 = hasAlpha8;
			else
				DEBUG_FATAL(hasAlpha8 != firstHasAlpha8, ("StaticMaterial::load_0001 some mip map levels have alph, some don't"));

			if (iff->seekChunk(TAG_PX32, Iff::FlagOptional))
			{
				hasPixel32 = true;

				if (!rgbMap32)
				{
					rgbMap32 = NEW Pixel32 *[mipMapLevels];
					memset(rgbMap32, 0, sizeof(*rgbMap32) * mipMapLevels);
				}

				DEBUG_FATAL(iff->chunkLength(sizeof(Pixel32)) != ((width >> i) * (height >> i)), ("StaticMaterial::load_0001 px32 bad size"));
				rgbMap32[i] = MEM_OWN(iff->read_pixel32(Iff::ReadRest));

				{
					int j;
					for (j = 0; j < (((width >> i) * (height >> i))); ++j)
						rgbMap32[i][j] = rgbMap32[i][j] ^ 0xff000000;
				}

				iff->exitChunk();
			}
			else
			{
				hasPixel32 = false;
			}

			if (i == 0)
				firstHasPixel32 = hasPixel32;
			else
				DEBUG_FATAL(hasPixel32 != firstHasPixel32, ("StaticMaterial::load_0001 some mip map levels have px32, some don't"));

		iff->exitForm();
	}

	DEBUG_FATAL(palette && !rgbMap8, ("StaticMaterial::load_0001 palette but no rgbMap8"));
}

// ----------------------------------------------------------------------

void StaticMaterial::load_BITM(Iff *iff)
{
	iff->seekForm(TAG_BITM, Iff::FlagOptional);
		iff->seekForm(TAG_FRAM);

			iff->seekChunk(TAG_INFO);
				width         = iff->read_int32();
				height        = iff->read_int32();
				if ((iff->read_int32() & 4) != 0)
					chromakey = ChromakeyOn;
			iff->exitChunk();

			// verify texture is square and power of two
			DEBUG_FATAL(width != height, ("StaticMaterial::load_BITM %s not square %dx%d", name, width, height));
			DEBUG_FATAL(!isPowerOfTwo(width), ("StaticMaterial::load_BITM %s not power of two %dx%d", name, width, height));
			DEBUG_FATAL(width > 256, ("StaticMaterial::load_BITM %s too big %dx%d", name, width, height));

			// load palette
			iff->seekForm(TAG_PAL);
				palette = PaletteList::fetch(*iff);
			iff->exitForm();

			// load the pixels
			iff->seekChunk(TAG_PXLS);
				DEBUG_FATAL(iff->chunkLength(sizeof(Pixel8)) != (width * height), ("StaticMaterial::load_BITM pxls bad size"));
				mipMapLevels = 1;
				rgbMap8 = NEW Pixel8*[mipMapLevels];
				rgbMap8[0] = MEM_OWN(iff->read_pixel8(Iff::ReadRest));
				color = ColorMap;
			iff->exitChunk();

			// load the optional alpha channel
			if (iff->seekChunk(TAG_ALPH, Iff::FlagOptional))
			{
				alpha = AlphaMap;

				if (iff->chunkLength(sizeof(Pixel8)) == 1)
					a = static_cast<u_int8>(255 - iff->read_u_int8());
				else
				{
				 	DEBUG_FATAL(iff->chunkLength(sizeof(Pixel8)) != (width * height), ("StaticMaterial::load_BITM alph bad size"));
					alphaMap = NEW Pixel8*[mipMapLevels];
					alphaMap[0] = MEM_OWN(iff->read_pixel8(Iff::ReadRest));

					{
						int j;
						for (j = 0; j < (width * height); ++j)
							alphaMap[0][j] = static_cast<u_int8>(255 - alphaMap[0][j]);
					}
				}

				iff->exitChunk();
			}

		iff->exitForm();
	iff->exitForm();
}

// ----------------------------------------------------------------------

void StaticMaterial::load_SALF(Iff *iff)
{
	alpha = AlphaMap;

	iff->seekForm(TAG_SALF);

		iff->seekChunk(TAG_INFO);
			width         = iff->read_int32();
			height        = iff->read_int32();
		iff->exitChunk();

		// verify texture is square and power of two
		DEBUG_FATAL(width != height, ("StaticMaterial::load_SALF %s not square %dx%d", name, width, height));
		DEBUG_FATAL(!isPowerOfTwo(width), ("StaticMaterial::load_SALF %s not power of two %dx%d", name, width, height));
		DEBUG_FATAL(width > 256, ("StaticMaterial::load_SALF %s too big %dx%d", name, width, height));

		// load the solid color rgb
		iff->seekChunk(TAG_COLR);
			r = static_cast<u_int8>(iff->read_int32());
			g = static_cast<u_int8>(iff->read_int32());
			b = static_cast<u_int8>(iff->read_int32());
		iff->exitChunk();

		// load the alpha channel
		iff->seekChunk(TAG_ALPH);
		 	DEBUG_FATAL(iff->chunkLength(sizeof(Pixel8)) != (width * height), ("StaticMaterial::load_SALF ALPH bad size"));
			mipMapLevels = 1;
			alphaMap = NEW Pixel8*[mipMapLevels];
			alphaMap[0] = MEM_OWN(iff->read_pixel8(Iff::ReadRest));

			{
				int j;
				for (j = 0; j < (width * height); ++j)
					alphaMap[0][j] = static_cast<u_int8>(255 - alphaMap[0][j]);
			}

		iff->exitChunk();

	iff->exitForm();

	// create the 32bit texture format
	rgbMap32    = NEW Pixel32*[mipMapLevels];
	rgbMap32[0] = NEW Pixel32[width * height];
	for (int i = 0; i < width * height; ++i)
		rgbMap32[0][i] = (alphaMap[0][i] << 24) | (r << 16) | (g << 8) || (b << 0);
}

// ----------------------------------------------------------------------

const StaticMaterial *StaticMaterial::getStaticMaterial(void) const
{
	return this;
}

// ----------------------------------------------------------------------

void  StaticMaterial::createData(void)
{
	DEBUG_FATAL(data, ("StaticMaterial::createData data already exists"));
	data = Gl::materialCreateData(this);
}

// ----------------------------------------------------------------------

void StaticMaterial::destroyData(void)
{
	DEBUG_FATAL(!data, ("StaticMaterial::createData data already exists"));
	Gl::materialDestroyData(data);
	data = NULL;
}

// ----------------------------------------------------------------------

Material *StaticMaterial::copy(void) const
{
	return Material::copy();
}

// ======================================================================

AnimatingMaterial::AnimatingMaterial(const char *newName, Iff *iff)
: Material(newName, MaterialIdAnimating)
{
	born = MaterialList::getTime();
	rate = CONST_REAL(16.0);
	maxFrame = 0;
	frame = NULL;
	load(iff);
}

// ----------------------------------------------------------------------

AnimatingMaterial::AnimatingMaterial(const AnimatingMaterial &am)
: Material(am)
{
	int i;

	// copy the animating material information
	born     = MaterialList::getTime();
	rate     = am.rate;
	maxFrame = am.maxFrame;
	frame    = NEW const Material*[maxFrame];

	// deep copy all the material frames
	for (i = 0; i < maxFrame; ++i)
		frame[i] = const_cast<Material *>(am.frame[i])->copy();
}

// ----------------------------------------------------------------------

AnimatingMaterial::~AnimatingMaterial(void)
{
	int i;

	for (i = 0; i < maxFrame; ++i)
	{
		MaterialList::release(frame[i]);
		frame[i] = NULL;
	}

	delete [] frame;
	frame = NULL;
}

// ----------------------------------------------------------------------

void AnimatingMaterial::load(Iff *iff)
{
	switch (iff->currentName())
	{
		case TAG_ANIM:
			load_ANIM(iff);
			break;

		case TAG_AMAT:
			iff->enterForm();
			switch (iff->currentName())
			{
				case TAG_0001:
					iff->enterForm();
					load_0001(iff);
					iff->exitForm();
					break;

				default:
					DEBUG_FATAL(true, ("AnimatingMaterial:: bad version"));
					break;
			}
			iff->exitForm();
			break;

		default:
			DEBUG_FATAL(true, ("AnimatingMaterial:: bad tag"));
			break;

	}
}

// ----------------------------------------------------------------------

void AnimatingMaterial::load_ANIM(Iff *iff)
{
	int i;

	iff->seekForm(TAG_ANIM);

		if (iff->seekForm(TAG_FRAM, Iff::FlagOptional))
		{
			maxFrame = iff->blockCount();
			frame = NEW const Material*[maxFrame];

			for (i = 0; i < maxFrame; ++i)
				frame[i] = MaterialList::fetch(iff);
			
			iff->exitForm();
		}
		else
		{
			iff->seekChunk(TAG_FRAM);

				maxFrame = iff->chunkLength(sizeof(int32));
				frame = NEW const Material*[maxFrame];

				for (i = 0; i < maxFrame; ++i)
					frame[i] = MaterialList::fetch(iff->read_int32());

			iff->exitChunk();
		}

		if (iff->seekChunk(TAG_HZ, Iff::FlagOptional))
		{
			rate = RECIP(iff->read_float());
			iff->exitChunk();
		}
		else
			if (iff->seekChunk(TAG_SEC, Iff::FlagOptional))
			{
				rate = iff->read_float();
				iff->exitChunk();
			}

	iff->exitForm();
}

// ----------------------------------------------------------------------

void AnimatingMaterial::load_0001(Iff *iff)
{
	int i;

	if (iff->seekForm(TAG_FRAM), Iff::FlagOptional)
	{
		maxFrame = iff->blockCount();
		frame = NEW const Material*[maxFrame];

		for (i = 0; i < maxFrame; ++i)
			frame[i] = MaterialList::fetch(iff);
		
		iff->exitForm();
	}
	else
	{
		iff->seekChunk(TAG_FRAM);

			maxFrame = iff->chunkLength(sizeof(int32));
			frame = NEW const Material*[maxFrame];

			for (i = 0; i < maxFrame; ++i)
				frame[i] = MaterialList::fetch(iff->read_int32());

		iff->exitChunk();
	}

	if (iff->seekChunk(TAG_HZ, Iff::FlagOptional))
	{
		rate = RECIP(iff->read_float());
		iff->exitChunk();
	}
	else
		if (iff->seekChunk(TAG_SEC, Iff::FlagOptional))
		{
			rate = iff->read_float();
			iff->exitChunk();
		}
}

// ----------------------------------------------------------------------

const StaticMaterial *AnimatingMaterial::getStaticMaterial(void) const
{
	return frame[getCurrentFrame()]->getStaticMaterial();
}

// ----------------------------------------------------------------------

Material *AnimatingMaterial::copy(void) const
{
	return NEW AnimatingMaterial(*this);
}

// ======================================================================
